#! /usr/bin/python3
# -*- coding: utf-8 -*-

#    Copyright (c) 2012 David Calle <davidc@framli.eu>

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import gettext
import locale
from gi.repository import GLib, GObject, Gio
from gi.repository import Accounts, Signon
from gi.repository import Dee
from gi.repository import Unity, Soup, SoupGNOME
import json
import time, datetime
import urllib


APP_NAME = "unity-lens-photos"
LOCAL_PATH = "/usr/share/locale/"

gettext.bindtextdomain(APP_NAME, LOCAL_PATH)
gettext.textdomain(APP_NAME)
_ = gettext.gettext

class Scope:

    def __init__(self, lens):
        self._scope = Unity.Scope.new("/net/launchpad/scope/photos/facebook")
        self._pending = []
        self._http = []
        for i in range(3):
            self._pending.append(None)
            self._http.append(self._get_http_session ())
        self._sources_options = []
        self._get_accounts_for_service ('facebook')
        self._authenticating = False
        self._enabled = False
        self._scope.connect("search-changed", self.on_search_changed)
        self._scope.connect("filters-changed", self.on_lens_active)
        self._scope.connect("notify::active", self.on_lens_active)
        self._scope.connect('preview-uri', self.on_preview_uri)
        self._scope.props.sources.connect("notify::filtering", self.on_filtering_changed)
        self._scope.export ()
        lens.add_local_scope (self._scope)


    def _get_http_session (self):
        session = Soup.SessionAsync()
        session.add_feature_by_type(SoupGNOME.ProxyResolverGNOME)
        return session


    def _get_accounts_for_service (self, service):
        self._accounts = []
        self._account_manager = Accounts.Manager.new_for_service_type("sharing")
        self._account_manager.connect("enabled-event", self._on_enabled_event);
        for account_service in self._account_manager.get_enabled_account_services():
            if account_service.get_account().get_provider_name() == service:
                self._add_account_service(account_service)


    def _add_account_service(self, account_service):
        for account in self._accounts:
            if account:
                if account.get_account_service() == account_service:
                    return
        self._accounts.append(self._account_to_login(account_service))
        print ('Added account %s' % (account_service))
        source_name = account_service.get_account().get_provider_name ().title()
        self._sources_options.append(source_name)
        self._scope.props.sources.add_option(source_name, source_name, None)


    def get_account_service(self):
        return self._account_service


    def _account_to_login(self,account_service):
        self._account_service = account_service
        self._account_service.connect("enabled", self._on_account_enabled)
        self._enabled = self._account_service.get_enabled()
        self._authenticating = False
        self._auth_token = None
        self._queued_search = None
        self._login()


    def _on_enabled_event(self, account_manager, account_id):
        account = self._account_manager.get_account(account_id)
        if account.get_provider_name() != "facebook": return
        for service in account.list_services():
            account_service = Accounts.AccountService.new(account, service)
            if account_service.get_enabled():
                self._add_account_service(account_service)


    def _add_account_service(self, account_service):
        for account in self._accounts:
            if account:
                if account.get_account_service() == account_service:
                    return
        self._accounts.append(self._account_to_login(account_service))
        print ('Added Facebook account %s' % (account_service))
        source_name = account_service.get_account().get_provider_name ().title()
        self._sources_options.append(source_name)
        self._scope.props.sources.add_option(source_name, source_name, None)


    def _login(self):
        if self._authenticating:
            return
        print ("Facebook : logging in")
        self._authenticating = True
        # Get the global account settings
        auth_data = self._account_service.get_auth_data()
        identity = auth_data.get_credentials_id()
        session_data = auth_data.get_parameters()
        self.auth_session = Signon.AuthSession.new(identity,
                auth_data.get_method())
        self.auth_session.process(session_data,
                auth_data.get_mechanism(),
                self._login_cb, None)


    def _login_cb(self, session, reply, error, user_data):
        print ("Facebook : login finished")
        self._authenticating = False
        if error:
#            print >> sys.stderr, "Got authentication error:", error.message
            return
        old_token = self._auth_token
#        if not reply.has_key("AccessToken") or not reply.has_key("TokenSecret"):
#            print >> sys.stderr, "Didn't find token in session: %s" % (reply,)
#            return        
        self._auth_token = reply["AccessToken"]
        if self._auth_token == old_token:
            print ("Got the same token")
            return
            
        if self._queued_search:
            self.on_search_changed(*self._queued_search)


    def _url_maker(self, cat, search_string, date):
        if date < 0:
            date = 180*86400
            now = int(time.time())
            date = now - date
            date_query = "created < '%s'" % date
        else:
            date_query = "created > '%s'" % date
        if search_string:
            if cat == 0:
                fql = {"q":"SELECT attachment, message, created_time FROM stream WHERE source_id = me() or source_id in (SELECT uid2 from friend where uid1=me ()) and type = 247 and strpos(caption,'%s') > 0 and %s LIMIT 500" % (search_string.lower (), date_query.replace("created", "created_time"))}
            elif cat == 1:
                fql = {"q":"SELECT object_id, src, link, caption, owner, created FROM photo WHERE aid in (SELECT aid FROM album where owner=me() ) and strpos(caption,'%s') > 0 and %s" % (search_string.lower (), date_query)}
            else:
                fql = {'q':'{"q1":"SELECT object_id, cover_object_id, location, link, name, description, owner FROM album where owner in (SELECT uid2 from friend where uid1=me ())","q2":"SELECT object_id, src, link, caption, owner, created FROM photo WHERE album_object_id IN (SELECT object_id FROM #q1) and strpos(caption,\'%s\') > 0 and %s LIMIT 100"}' % (search_string.lower (), date_query)}
        else:
            if cat == 0:
                fql = {"q":"SELECT attachment, message, created_time FROM stream WHERE source_id = me() or source_id in (SELECT uid2 from friend where uid1=me ()) and type = 247 and %s LIMIT 500" % date_query.replace("created", "created_time")}
            elif cat == 1:
                fql = {"q":"SELECT object_id, src, link, caption, owner, created FROM photo WHERE aid in (SELECT aid FROM album where owner=me() ) and %s" % date_query}
            else:
                fql = {'q':'{"q1":"SELECT object_id, cover_object_id, location, link, name, description, owner FROM album where owner in (SELECT uid2 from friend where uid1=me ())","q2":"SELECT object_id, src, link, caption, owner, created FROM photo WHERE album_object_id IN (SELECT object_id FROM #q1) and %s LIMIT 100"}' % date_query}
        url = 'https://graph.facebook.com/fql?%s&access_token=%s' % (urllib.parse.urlencode(fql), self._auth_token)
        return url


    def _on_account_enabled (self, account, enabled):
        self._enabled = enabled


    def on_filtering_changed(self, *_):
        """Run another search when a filter change is notified."""
        for source in self._sources_options:
            filtering = self._scope.props.sources.props.filtering
            active = self._scope.props.sources.get_option(source).props.active
            if (active and filtering) or (not active and not filtering):
                if not self._enabled:
                    self._enabled = True
                    self._scope.queue_search_changed(Unity.SearchType.DEFAULT)
            else:
                self.cancel_all_searches ()
            print ("    %s enabled : %s" % (source, self._enabled))


    def cancel_all_searches (self):
        self._enabled = False
        for i in range(len(self._pending)):
            try:
                self._http[i].cancel_message(self._pending[i],Soup.KnownStatusCode.CANCELLED)
            except:
                pass
        self._scope.props.results_model.clear ()


    def on_lens_active(self, *_):
        """ Update results when the lens is opened """
#        if self._scope.props.active:
        self._scope.queue_search_changed(Unity.SearchType.DEFAULT)


    def on_preview_uri(self, scope, uri):
        """Preview request handler"""
        preview = None
        model = self._scope.props.results_model
        iter = model.get_first_iter()
        end_iter = model.get_last_iter()
        while iter != end_iter:
            if model.get_value(iter, 0) == uri:
                photo_id = model.get_value(iter, 5).split("_ulp-date_")[0]
                meta = self.getMetadataForPhoto (photo_id,model.get_value(iter, 2))
                location, date, image, title, description, album, position, album_size = None, None, None, '', '', None, None, None
                if meta[6]:
                    title = meta[6]
                elif meta[5]:
                    title = meta[5]
                else:
                    title = model.get_value(iter, 4)
                if meta[1]:
                    description = meta[1]
                preview = Unity.GenericPreview.new(title, description, None)
                if meta[2]:
                    preview.props.image_source_uri = meta[2]
                else:
                    preview.props.image_source_uri = model.get_value(iter, 1)
                if meta[8]:
                    preview.props.subtitle = meta[8]
                if meta[7]:
                    preview.add_info(Unity.InfoHint.new("location", _("Location"), None, meta[7]))
                if meta[9]:
                    album = meta[9]
                    if meta[10] and meta[11]:
                        album = "%s, %i of %i" % (meta[9], meta[10], meta[11])
                    preview.add_info(Unity.InfoHint.new("album", _("Album"), None, album))
                if meta[4]:
                    preview.add_info(Unity.InfoHint.new("with", _("With"), None, meta[4]))

                gfile_icon = Gio.file_new_for_path("/usr/share/unity-lens-photos/media/Facebook.svg")
                gicon = Gio.FileIcon.new (gfile_icon)
                view_action = Unity.PreviewAction.new("view", _("View"), gicon)
                view_action.connect('activated', self.on_activate_uri)
                preview.add_action(view_action)
                break
            iter = model.next(iter)
        if preview == None:
            print ("Couldn't find model row for requested preview uri: '%s'", uri)
        return preview


    def getMetadataForPhoto(self, oid, cat):
        if cat == 0:
            fql = {'q':'{"query1":"SELECT images, owner, link, caption, created, place_id, album_object_id, position FROM photo WHERE pid=\'%s\'","query2":"SELECT subject, text FROM photo_tag WHERE pid=\'%s\'","query3":"SELECT username, name FROM user WHERE uid IN (SELECT owner FROM #query1)","query4":"SELECT name FROM place WHERE page_id IN (SELECT place_id FROM #query1)","query5":"SELECT name, size FROM album WHERE object_id IN (SELECT album_object_id FROM #query1)"}' % (oid, oid)}
        else:
            fql = {'q':'{"query1":"SELECT images, owner, link, caption, created, place_id, album_object_id, position FROM photo WHERE object_id=%s","query2":"SELECT subject, text FROM photo_tag WHERE object_id=%s","query3":"SELECT username, name FROM user WHERE uid IN (SELECT owner FROM #query1)","query4":"SELECT name FROM place WHERE page_id IN (SELECT place_id FROM #query1)","query5":"SELECT name, size FROM album WHERE object_id IN (SELECT album_object_id FROM #query1)"}' % (oid, oid)}
        url = 'https://graph.facebook.com/fql?%s&access_token=%s' % (urllib.parse.urlencode(fql), self._auth_token)
        raw_results = urllib.request.urlopen(url).read ()
        meta = json.loads(raw_results.decode("utf8"))

        link, caption, image, date = None, None, None, None
        subject, tags, album, position = None, None, None, None
        place, name, username, album_size = None, None, None, None
        tag_list = []
        
        for m in meta['data'][0]['fql_result_set']:
            date = datetime.datetime.fromtimestamp(m['created']).strftime('%d %b %Y')
            link = m['link']
            caption = m['caption']
            position = m['position']
            image = m['images'][0]['source']
        for m in meta['data'][1]['fql_result_set']:
            if m['text'] not in tag_list:
                tag_list.append(m['text'])
            tags = ", ".join(tag_list)
        for m in meta['data'][2]['fql_result_set']:
            username = m['username']
            name = m['name']
        for m in meta['data'][3]['fql_result_set']:
            place = m['name']
        for m in meta['data'][4]['fql_result_set']:
            album = m['name']
            album_size = m['size']

        return [link,caption, image, subject, tags, username, name, place, date, album, position, album_size]


    def on_activate_uri (self, *_):
        return


    def check_date_filter(self):
        """Get active option for a filter name"""
        try:
            date = self._scope.get_filter("date").get_active_option().props.id
            date = int(date)*86400
            now = int(time.time())
            date = now - date
        except (AttributeError):
            date = 0
        return date


    def on_search_changed (self, scope, search, search_type, cancellable):
        model = search.props.results_model
        model.clear()
        self._queued_search = (scope, search, search_type, cancellable)
        if self._authenticating:
            print ("authenticating, queuing search")
            return
        search_string = search.props.search_string.strip()
        print ('Facebook : new search %s' % search_string)
        if self._enabled:
            if search_type is Unity.SearchType.DEFAULT:
                date = self.check_date_filter ()
                if search_string:
                    for i in range(1,3):
                        if self._pending[i] is not None:
                            self._http[i].cancel_message(self._pending[i],
                                                        Soup.KnownStatusCode.CANCELLED)
                        self._pending[i] = Soup.Message.new("GET", self._url_maker(i, search_string, date))
                        self._http[i].queue_message(self._pending[i],self._search_cb,[search_string, model, i, search])
                else:
                    for i in range(3):
                        if self._pending[i] is not None:
                            self._http[i].cancel_message(self._pending[i],
                                                        Soup.KnownStatusCode.CANCELLED)
                        self._pending[i] = Soup.Message.new("GET", self._url_maker(i, search_string, date))
                        self._http[i].queue_message(self._pending[i],self._search_cb,[search_string, model, i, search])



    def update_results_model(self, search, model, results, cat):
        date = self.check_filters("date")
        counter = 0
        if len(results) > 0:
            for photo in results['data']:
                if counter < 100:
                    if cat == 0:
                        timestamp = photo['created_time']
                        try:
                            for media in photo['attachment']['media']:
                                counter += 1
                                title = media['alt']
                                comment = str(media['photo']['pid'])+"_ulp-date_"+str(timestamp)
                                uri = media['href']
                                icon_hint = media['src']
                                model.append (uri, icon_hint, cat,"text/html", title, comment, uri)
                                model.flush_revision_queue ()
                        except:
                            pass
                    elif cat == 2:
                        if photo['name'] == 'q2':
                            for p in photo['fql_result_set']:
                                counter += 1
                                timestamp = p['created']
                                title = p['caption']
                                comment = str(p['object_id'])+"_ulp-date_"+str(timestamp)
                                uri = p['link']
                                icon_hint = p['src']
                                model.append (uri, icon_hint, cat,"text/html", title, comment, uri)
                                model.flush_revision_queue ()
                    else:
                        counter += 1
                        timestamp = photo['created']
                        title = photo['caption']
                        comment = str(photo['object_id'])+"_ulp-date_"+str(timestamp)
                        uri = photo['link']
                        icon_hint = photo['src']
                        model.append (uri, icon_hint, cat,"text/html", title, comment, uri)
                        model.flush_revision_queue ()
        print ('Facebook : Added %i results to category %i' % (counter, cat))


    def check_filters(self, filter_name):
        try:
            filter = self._scope.get_filter(filter_name).get_active_option().props.id
        except (AttributeError):
            filter = None
        return filter


    def _search_cb(self, session, msg, search_args):
        results = self._handle_search_msg(msg, search_args[2])
        self.update_results_model(search_args[0], search_args[1], results, search_args[2])
        if search_args[3] and search_args[2] == 3:
            Unity.LensSearch.finished(search_args[3])


    def _handle_search_msg(self, msg, cat):
        results = []
        if msg.status_code == 401:
            print ('Facebook : Retrying for category %s' % cat)
            self._scope.queue_search_changed(Unity.SearchType.DEFAULT)
        elif msg.status_code != 200:
            self._pending[cat] = None
            print ("Error: Unable to get results from the server")
            print ("       %d: %s" % (msg.status_code, msg.reason_phrase))
        else:
            self._pending[cat] = None
            try:
                results = json.loads(msg.response_body.data)
            except ValueError:
                print ("Error: got garbage json from the server")
        return results
